package com.brackeen.evolution.core.game;

import com.brackeen.evolution.core.Actor;
import com.brackeen.evolution.core.Evolution;
import com.brackeen.evolution.core.Scene;
import com.brackeen.evolution.core.util.BezierPath;
import com.brackeen.evolution.core.util.MathUtil;
import playn.core.Image;
import playn.core.ImageLayer;
import playn.core.Layer;
import playn.core.PlayN;
import pythagoras.f.FloatMath;
import pythagoras.f.Point;

public class Alien extends Actor {

    public enum State {
        MOVING_TO_FORMATION,
        IN_FORMATION,
        ATTACKING,
    }
    
    private final AlienGenome genome;
    private AlienArmada armada;
    private Point armadaPosition;
            
    private State state;
    private float durationInState;
    private float stateStartX;
    private float stateStartY;
    private boolean firstMoveIn;
    
    private int health;
    
    // Attack mode
    private Spaceship attackTarget;
    private BezierPath path;
    private float timeUntilCanFire;
    
    public Alien(AlienGenome genome) {
        this.genome = genome;
        
        Image image = genome.getImage();
        ImageLayer layer = PlayN.graphics().createImageLayer(image);
        layer.setSize(image.width(), image.height());
        layer.setOrigin(image.width()/2, image.height()/2);
        setLayer(layer);

        setState(State.IN_FORMATION);
        
        health = (int)(Math.pow(genome.getShields() / 100.0f, 2) * 100);
        
        setRested();
        
        firstMoveIn = true;
        
    }

    public void setArmada(AlienArmada armada) {
        this.armada = armada;
    }
    
    public void setState(State state) {
        if (this.state != state) {
            this.state = state;
            durationInState = 0;
            stateStartX = layer().transform().tx();
            stateStartY = layer().transform().ty();
            timeUntilCanFire = 0;
        }
    }

    public State getState() {
        return state;
    }

    public int getHealth() {
        return health;
    }
    
    public float getDurationInState() {
        return durationInState;
    }
    
    public float getVelocity() {
        return genome.getSpeed() * 5;
    }
    
    public float getRestTime() {
        return Math.max(1, genome.getRestTime() / 10.0f);
    }
    
    public boolean isRested() {
        return (durationInState >= getRestTime());
    }
    
    private void setRested() {
        durationInState = getRestTime();
    }
    
    public AlienGenome getGenome() {
        return genome;
    }
    
    public void setArmadaPosition(float x, float y) {
        armadaPosition = new Point(x, y);
    }
    
    public Point getArmadaPosition() {
        return armadaPosition;
    }
    
    public void doDamage(float damage) {
        health -= damage;
        if (health <= 0) {
            setShouldRemove(true);
            Evolution.playSound("audio/explode");
        }
        else {
            Evolution.playSound("audio/hit");
        }
    }
    
    @Override
    public void update(Evolution evolution, float delta) {
        float elapsedTime = delta / 1000.0f;
        durationInState += elapsedTime;

        if (timeUntilCanFire > 0) {
            timeUntilCanFire -= elapsedTime;
        }

        if (state == State.ATTACKING) {
            float t = durationInState;
            float v = getVelocity();
            float d = path.getLength();
            float p = v * t / d;
            if (p >= 1) {
                setState(State.MOVING_TO_FORMATION);
                p = 1;
            }
            float x = path.getX(p);
            float y = path.getY(p);
            layer().setTranslation(x, y);
            layer().setRotation(path.getAngle(p) + FloatMath.PI / 2);

            if (timeUntilCanFire <= 0 && armada != null && armada.isPastFrontLine(this)) {
                if (y >= attackTarget.layer().transform().ty() - attackTarget.layer().height()) {
                    // Past the player - don't fire again
                    timeUntilCanFire = 10000;
                }
                else {
                    timeUntilCanFire = (150 - genome.getFireSpeed()) / 150.0f;
                    //System.out.println(timeUntilCanFire);

                    Scene scene = getScene();
                    if (scene != null) {
                        Image image = genome.getFireDamage() >= 75 ? armada.getLargeBulletImage() : armada.getBulletImage();
                        Bullet bullet = new Bullet(Bullet.Source.ALIEN, image, x, y, getVelocity() * 1.5f, genome.getFireDamage());
                        scene.addActor(bullet);
                    }
                }
            }
        }
        else if (state == State.MOVING_TO_FORMATION && armada != null) {
            Point pos = armada.getViewLocation(getArmadaPosition());
            float t = durationInState;
            float v = getVelocity();
            float d = MathUtil.dist(stateStartX, stateStartY, pos.x, pos.y);
            float p = v * t / d;
            if (p >= 1) {
                setState(State.IN_FORMATION);
                p = 1;
                if (firstMoveIn) {
                    firstMoveIn = false;
                    setRested();
                }
            }
            layer().setTranslation(
                    stateStartX + p * (pos.x - stateStartX),
                    stateStartY + p * (pos.y - stateStartY));
            if (!firstMoveIn) {
                rotateToAngle(FloatMath.atan2(pos.y - stateStartY, pos.x - stateStartX) + FloatMath.PI / 2);
            }
        }
        else if (state == State.IN_FORMATION && armada != null) {
            Point p = armada.getViewLocation(getArmadaPosition());
            layer().setTranslation(p.x, p.y);
            rotateToAngle(0);
        }
    }

    private void rotateToAngle(float angle) {
        float fromAngle = layer().transform().rotation();
        if (fromAngle != angle) {
            float toAngle;
            float v = getVelocity() / 10000;
            if (Math.abs(fromAngle - angle) <= v) {
                toAngle = angle;
            }
            else if (angle < fromAngle) {
                toAngle = fromAngle - v;
            }
            else {
                toAngle = fromAngle + v;
            }
            layer().setRotation(toAngle);
        }
    }

    public void attack(Spaceship spaceship) {
        if (spaceship == null || spaceship.shouldRemove()) {
            return;
        }
        this.attackTarget = spaceship;
        Layer.HasSize l = layer();
        float x = l.transform().tx();
        float y = l.transform().ty();
        float x2 = x;
        float y2 = spaceship.layer().transform().ty();
                
        float d = l.width() * 4;
        float s;
        if (getArmadaPosition().x < armada.getWidth() /2) {
            s = d;
        }
        else {
            s = -d;
        }
        
        if (genome.getAcceleration() >= 50) {
            path = new BezierPath(
                    new float[] { BezierPath.MOVE_TO, x, y },
                    new float[] { BezierPath.CURVE_TO, x, y - d, x + s, y - d, x + s, y },
                    new float[] { BezierPath.CURVE_TO, x + s, y + d, x2, y2 - d, x2, y2 },
                    new float[] { BezierPath.CURVE_TO, x2, y2 + d, x2 + s, y2 + d, x2 + s, y2 },
                    new float[] { BezierPath.CURVE_TO, x2 + s, y2 - d, x2, y2 - d, x2, y2 },
                    new float[] { BezierPath.CURVE_TO, x2, y2 + d, x2 + s, y2 + d, x2 + s, y2 }
                    );            
        }
        else {
            path = new BezierPath(
                    new float[] { BezierPath.MOVE_TO, x, y },
                    new float[] { BezierPath.CURVE_TO, x, y - d, x + s, y - d, x + s, y },
                    new float[] { BezierPath.CURVE_TO, x + s, y + d, x2, y2 - d, x2, y2 },
                    new float[] { BezierPath.CURVE_TO, x2, y2 + d, x2 + s, y2 + d, x2 + s, y2 }
                    );
        }
        setState(State.ATTACKING);
    }
}
