package com.brackeen.evolution.core.game;

import com.brackeen.evolution.core.Evolution;
import com.brackeen.evolution.core.util.MathUtil;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import playn.core.Image;
import pythagoras.f.Point;

/*
Immutable after constructor is called.
*/
public class AlienGenome {
    
    private static final String BASE_64_CHARS = 
            "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_";
    private static final int WIDTH = SpriteCreator.WIDTH;
    private static final int HEIGHT = SpriteCreator.HEIGHT;
    
    // Indexes to Evolution.COLORS
    private static final int[][] DEFAULT_GRID = {
        { 0, 0, 0, 0, 0, 0, 0, 0, },
        { 0, 0, 0, 0, 0, 0, 0, 0, },
        { 0, 0, 0, 0, 0, 0, 0, 0, },
        { 0, 0, 0, 0, 0, 0, 0, 0, },
        { 0, 0, 0, 0, 0, 0, 0, 6, },
        { 0, 0, 0, 0, 0, 6, 6, 6, },
        { 0, 0, 0, 0, 6, 6, 6, 6, },
        { 0, 0, 0, 0, 6, 6, 2, 6, },
        { 0, 0, 0, 0, 6, 6, 6, 6, },
        { 0, 0, 0, 0, 3, 6, 6, 6, },
        { 0, 0, 0, 3, 3, 3, 6, 6, },
        { 0, 0, 0, 3, 3, 3, 3, 0, },
        { 0, 0, 0, 0, 3, 3, 0, 0, },
        { 0, 0, 0, 0, 0, 0, 0, 0, },
        { 0, 0, 0, 0, 0, 0, 0, 0, },
        { 0, 0, 0, 0, 0, 0, 0, 0, },
    };
    
    private int[][] grid;
    private Image image;
    private int age;
        
    // Abilities, from 0 to 100
    private int speed;
    private int acceleration;
    private int fireDamage;
    private int fireSpeed;
    private int shields;
    private int restTime;
    
    public AlienGenome() {
        this(DEFAULT_GRID, 0, false);
    }
    
    public AlienGenome(String importString) {
        this(importString(importString), 0, false);
    }
    
    private AlienGenome(int[][] grid, int age, boolean mutate) {
        this.grid = grid;
        if (mutate) {
            this.age = age + doMutate(age);
        }
        else {
            this.age = age;
        }
        this.image = SpriteCreator.create(grid);
        
        // Stats
        int[] colorCount = new int[Evolution.COLORS.length];
        float[] pColorCount = new float[Evolution.COLORS.length];
        int numPixels = 0;
        int numPixelsInFront = 0;
        int numPixelsInBack = 0;

        for (int y = 0; y < HEIGHT; y++) {
            for (int x = 0; x < WIDTH; x++) {
                int colorIndex = get(x, y);
                colorCount[colorIndex]++;
                if (colorIndex != 0) {
                    numPixels++;
                    if (y < HEIGHT/2) {
                        numPixelsInFront++;
                    }
                    else {
                        numPixelsInBack++;
                    }
                }
            }
        }
        if (numPixels > 0) {
            for (int i = 0; i < Evolution.COLORS.length; i++) {
                pColorCount[i] = (float)colorCount[i] / numPixels;
            }
        }
        float pNumPixels = (float)numPixels / (WIDTH * HEIGHT);
        float pNumPixelsInFront = (float)numPixelsInFront / numPixels;
     
        speed = 50;
        fireDamage = 50;
        fireSpeed = 50;
        shields = 50;
        acceleration = 50;
        restTime = 50;

        // Colors
        speed += 50 * pColorCount[1]; // White
        fireSpeed += 50 * pColorCount[2]; // Red
        fireDamage += 25 * pColorCount[3]; // Blue
        restTime -= 25 * pColorCount[3]; // Blue
        shields += 25 * pColorCount[4]; // Green
        restTime -= 25 * pColorCount[4]; // Green
        fireDamage += 50 * pColorCount[5]; // Purple
        shields += 50 * pColorCount[6]; // Yellow
        acceleration += 50 * pColorCount[7]; // Cyan
        
        // Num Pixels
        int p;
        if (pNumPixels < 0.20) {
            p = -20;
        }
        else if (pNumPixels > 0.40) {
            p = 20;
        }
        else {
            p = (int)((pNumPixels - 0.20) * 200 - 20);
        }
        shields += p;
        acceleration -= p;
        fireDamage += p;
        restTime += p;
        
        // Front/back distribution
        fireSpeed += 40 * (pNumPixelsInFront - 0.5);
        speed -= 40 * (pNumPixelsInFront - 0.5);
        shields += 40 * (pNumPixelsInFront - 0.5);
   
    }

    public Image getImage() {
        return image;
    }

    public int getSpeed() {
        return speed;
    }

    public int getFireDamage() {
        return fireDamage;
    }

    public int getFireSpeed() {
        return fireSpeed;
    }

    public int getShields() {
        return shields;
    }

    public int getAcceleration() {
        return acceleration;
    }
    
    public int getRestTime() {
        return restTime;
    }
    
   
    
    public String getExportString() {
        assert(Evolution.COLORS.length == 8);
        
        String s = "";
        for (int y = 0; y < HEIGHT; y++) {
            for (int x = 0; x < WIDTH; x+=2) {
                int i = get(x, y) | (get(x+1,y) << 3);
                s += BASE_64_CHARS.charAt(i);
            }
        }
        return s;
    }
    
    private static int[][] importString(String s) {
        if (s == null) {
            return DEFAULT_GRID;
        }
        int n = WIDTH * HEIGHT / 2;
        if (s.length() != n) {
            System.out.println("Import string not " + n + " characters long.");
            return DEFAULT_GRID;
        }
        int[][] newGrid = new int[HEIGHT][WIDTH];
        int i = 0;
        for (int y = 0; y < HEIGHT; y++) {
            for (int x = 0; x < WIDTH; x+=2) {
                char ch = s.charAt(i++);
                int index = BASE_64_CHARS.indexOf(ch);
                if (index == -1) {
                    System.out.println("Illegal character in import string: " + ch);
                    return DEFAULT_GRID;
                }
                newGrid[y][x] = index & 0x7;
                newGrid[y][x+1] = (index >> 3) & 0x07;
            }
        }
        return newGrid;
    }
    
    private int get(int x, int y) {
        if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT) {
            return 0;
        }
        else {
            return grid[y][x];
        }
    }

    @Override
    public String toString() {
        return getDescription();
    }
    
    public String getDescription() {
//        String gridString = "{\n";
//        for (int y = 0; y < HEIGHT; y++) {
//            gridString += "    { ";
//            for (int x = 0; x < WIDTH; x++) {
//                gridString += grid[y][x] + ", ";
//            }
//            gridString += "}\n";
//        }
//        gridString += "}";
        return "AlienGenome " + 
                "genome = " + getExportString() + "\n" + 
                "age = " + age + "\n" +
                "speed = " + speed + "\n" +
                "acceleration = " + acceleration + "\n" +
                "restTime = " + restTime + "\n" +
                "fireDamage = " + fireDamage + "\n" +
                "fireSpeed = " + fireSpeed + "\n" +
                "shields = " + shields + "\n";
    }
    
    //
    // Mutation
    //
    
    private void debugPrint(String statement) {
        //System.out.println(statement);
    }
    
    public AlienGenome mutatedCopy() {
        // Copy grid
        int[][] newGrid = new int[HEIGHT][WIDTH];
        for (int i = 0; i < HEIGHT;  i++) {
            System.arraycopy(grid[i], 0, newGrid[i], 0, WIDTH);
        }
        return new AlienGenome(newGrid, age, true);
    }
    
    private int doMutate(int age) {
        int iterations;
        if (age < 64) {
            iterations = 8;
        }
        else if (age < 1024) {
            iterations = 16;
        }
        else {
            iterations = 32;
        }
        int changeCount = 0;
        while (changeCount < iterations) {
            if (mutate()) {
                changeCount++;
            }
        }
        return iterations;
    }
    
    private void set(int x, int y, int color) {
        if (x >= 0 && y >= 0 && x < WIDTH && y < HEIGHT) {
            grid[y][x] = color;
        }
    }
    
    private int numNeighborsOfColor(int x, int y, int color) {
        return                             
                ((get(x-1, y-1) == color) ? 1 : 0) +
                ((get(x, y-1) == color) ? 1 : 0) +
                ((get(x+1, y-1) == color) ? 1 : 0) +
                ((get(x-1, y) == color) ? 1 : 0) +
                ((get(x+1, y) == color) ? 1 : 0) +
                ((get(x-1, y+1) == color) ? 1 : 0) +
                ((get(x, y+1) == color) ? 1 : 0) +
                ((get(x+1, y+1) == color) ? 1 : 0);
    }
    
    private int mostPopularNeighborColor(int xx, int yy, boolean includeBlack) {
        int[] colorCount = new int[Evolution.COLORS.length];
        for (int y = yy-1; y <= yy+1; y++) {
            for (int x = xx-1; x <= xx+1; x++) {
                if (x != xx || y != yy) {
                    colorCount[get(x, y)]++;
                }
            }
        }
        int maxIndex = includeBlack ? 0 : 1;
        for (int i = maxIndex + 1; i < colorCount.length; i++) {
            if (colorCount[i] >= colorCount[maxIndex]) {
                maxIndex = i;
            }
        }
        return maxIndex;
    }
    
    private int maxEnemyCount(int xx, int yy, boolean includeBlack) {
        int[] colorCount = new int[Evolution.COLORS.length];
        for (int y = yy-1; y <= yy+1; y++) {
            for (int x = xx-1; x <= xx+1; x++) {
                if (x != xx || y != yy) {
                    colorCount[get(x, y)]++;
                }
            }
        }
        int maxIndex = includeBlack ? 0 : 1;
        for (int i = maxIndex + 1; i < colorCount.length; i++) {
            if (i != get(xx, yy) && colorCount[i] >= colorCount[maxIndex]) {
                maxIndex = i;
            }
        }
        return colorCount[maxIndex];
    }

    private boolean mutate() {
        
        boolean somethingChanged = false;
        
        // Count colors
        int[] colorCount = new int[Evolution.COLORS.length];
        int pixelCount = 0;
        HashSet<Integer> unusedColors = new HashSet<Integer>();
        for (int i = 1; i < Evolution.COLORS.length; i++) {
            unusedColors.add(i);
        }
        HashSet<Integer> usedColors = new HashSet<Integer>();
        for (int y = 0; y < HEIGHT; y++) {
            for (int x = 0; x < WIDTH; x++) {
                int colorIndex = get(x, y);
                if (colorIndex != 0) {
                    usedColors.add(colorIndex);
                    pixelCount++;
                }
                colorCount[colorIndex]++;
            }
        }
        for (Integer usedColor : usedColors) {
            unusedColors.remove(usedColor);
        }
        float amountSolid = (float)pixelCount / (WIDTH * HEIGHT);
        
        // Find "lonely" pixels - 0 or 1 neighbors of the same color.
        ArrayList<Point> lonelyPixels = new ArrayList<Point>();
        for (int y = 0; y < HEIGHT; y++) {
            for (int x = 0; x < WIDTH; x++) {
                int c = get(x, y);
                int friendCount = numNeighborsOfColor(x, y, c);
                if (friendCount <= 1) {
                    lonelyPixels.add(new Point(x, y));
                }
            }
        }
        debugPrint("Used colors: " + usedColors.size() + " amountSolid: " + amountSolid + " lonelyPixels: " + lonelyPixels.size());
        
        // Mutation 1: change colors
        if (MathUtil.probability(0.25) && usedColors.size() >= 2) {
            int fromColor = MathUtil.random(usedColors);
            usedColors.remove(fromColor);
            
            boolean choseExisting = 
                    (usedColors.size() > 0 && MathUtil.probability(0.05)) || 
                    (usedColors.size() > 3 && MathUtil.probability(0.25));
            
            int toColor;
            if (!choseExisting && unusedColors.size() > 0) {
                toColor = MathUtil.random(unusedColors);
                unusedColors.remove(toColor);
            }
            else {
                toColor = MathUtil.random(usedColors);
            }
            
            usedColors.add(toColor);
            unusedColors.add(fromColor);
                        
            for (int y = 0; y < HEIGHT; y++) {
                for (int x = 0; x < WIDTH; x++) {
                    int colorIndex = get(x, y);
                    if (colorIndex == fromColor) {
                        set(x, y, toColor);
                    }
                }
            }
            
            int fromColorCount = colorCount[fromColor];
            colorCount[fromColor] = 0;
            colorCount[toColor] += fromColorCount;
            debugPrint("Change color");
            somethingChanged = true;
        }
        
        // Mutation 2: change color of one pixel
        if ((amountSolid > 0.5 || MathUtil.probability(0.25)) && usedColors.size() > 0) {
            List<Point> positions = new ArrayList<Point>();
            boolean tooHigh = false;
            int oldColor;
            if (amountSolid < 0.2) {
                oldColor = 0;
            }
            else if (amountSolid > 0.4) {
                oldColor = MathUtil.random(usedColors);
                tooHigh = true;
            }
            else if (MathUtil.probability(0.4)) {
                oldColor = MathUtil.random(usedColors);
            }
            else {
                oldColor = 0;
            }
            
            if (tooHigh) {
                for (int y = 0; y < HEIGHT; y++) {
                    for (int x = 0; x < WIDTH; x++) {
                        int c = get(x, y);
                        if (c != 0 && (x == 0 || y == 0 || y == HEIGHT - 1)) {
                            positions.add(new Point(x, y));
                        }
                    }
                }
                if (positions.isEmpty()) {
                    for (int y = 0; y < HEIGHT; y++) {
                        for (int x = 0; x < WIDTH; x++) {
                            boolean edge = x == 0 || y == 0 || x == WIDTH - 1 || y == HEIGHT - 1;
                            int c = get(x, y);
                            
                            if (c == oldColor && numNeighborsOfColor(x, y, 0) >= 2 + (edge ? 3 : 0)) {
                                positions.add(new Point(x, y));
                            }
                        }
                    }
                }
            }
            else {
                // First, find a pixel with 4 enemy neigbors
                for (int y = 0; y < HEIGHT; y++) {
                    for (int x = 0; x < WIDTH; x++) {
                        int c = get(x, y);
                        if (c == oldColor && maxEnemyCount(x, y, false) >= 4) {
                            positions.add(new Point(x, y));
                        }
                    }
                }
                if (positions.isEmpty()) {
                    for (int y = 0; y < HEIGHT; y++) {
                        for (int x = 0; x < WIDTH; x++) {
                            int c = get(x, y);
                            if (c == oldColor && (
                                get(x, y-1) != c ||
                                get(x-1, y) != c ||
                                get(x+1, y) != c ||
                                get(x, y+1) != c)) {

                                positions.add(new Point(x, y));
                            }
                        }
                    }
                }
            }
            
            if (positions.size() > 0) {
                Point p = MathUtil.random(positions);
                int x = (int)p.x;
                int y = (int)p.y;
                int c = get(x, y);
                int newColor;
                if (tooHigh) {
                    newColor = 0;
                }
                else {
                    newColor = mostPopularNeighborColor(x, y, false);
                }
                boolean cancel = newColor == 0 && numNeighborsOfColor(x, y, 0) < 3;
                if (!cancel) {
                    debugPrint("Change to " + newColor);
                    set(x, y, newColor);

                    colorCount[c]--;
                    colorCount[newColor]++;
                    if (c != 0 && colorCount[c] == 0) {
                        usedColors.remove(c);
                        unusedColors.add(c);
                    }
                    somethingChanged = true;
                }
            }
        }
        
        // Mutation 3: Add new color
        if (lonelyPixels.size() < 2 && MathUtil.probability(0.1) && unusedColors.size() > 0 && usedColors.size() < 5) {
            List<Point> positions = new ArrayList<Point>();
            
            // Try to find one pixel surrounded by one color.
            for (int y = 1; y < HEIGHT-1; y++) {
                for (int x = 1; x < WIDTH-1; x++) {
                    int c = get(x, y);
                    if (c != 0 && numNeighborsOfColor(x, y, c) == 8) {
                        positions.add(new Point(x, y));
                    }
                }
            }
            
            // Replace a black color that is next to at least 3 of one color, and at least 3 black
            for (int y = 1; y < HEIGHT-1; y++) {
                for (int x = 1; x < WIDTH-1; x++) {
                    int c = get(x, y);
                    if (c == 0) {
                        int blackCount = numNeighborsOfColor(x, y, c);
                        int neighborColor = mostPopularNeighborColor(x, y, false);
                        int neighborColorCount = numNeighborsOfColor(x, y, neighborColor);
                        if (blackCount >= 3 && neighborColorCount >= 3 && blackCount + neighborColorCount == 8) {
                            positions.add(new Point(x, y));
                        }
                    }
                }
            }
            
            if (positions.size() > 0) {
                Point p = MathUtil.random(positions);
                int x = (int)p.x;
                int y = (int)p.y;
                int newColor = MathUtil.random(unusedColors);
                int oldColor = get(x, y);
                set(x, y, newColor);
                
                debugPrint("Add new: " + newColor);

                unusedColors.remove(newColor);
                usedColors.add(newColor);
                colorCount[oldColor] --;
                colorCount[newColor] ++;
                if (oldColor != 0 && colorCount[oldColor] == 0) {
                    usedColors.remove(oldColor);
                    unusedColors.add(oldColor);
                }
                somethingChanged = true;
            }
        }
        else if (lonelyPixels.size() >= 2 && MathUtil.probability(0.25)) {
            Point p = MathUtil.random(lonelyPixels);
            int x = (int)p.x;
            int y = (int)p.y;
            int oldColor;
            int newColor;
            boolean isVeryLonely = numNeighborsOfColor(x, y, get(x, y)) == 0;
            boolean hasMany = lonelyPixels.size() >= 3;
            if ((hasMany && isVeryLonely) || (isVeryLonely && MathUtil.probability(0.5))) {
                oldColor = get(x, y);
                newColor = mostPopularNeighborColor(x, y, false);
                set(x, y, newColor);
                debugPrint("Kill Lonely Pixel: " + newColor);
            }
            else {
                int nx;
                int ny;
                while (true) {
                    nx = x + MathUtil.random(-1, 2);
                    ny = y + MathUtil.random(-1, 2);
                    if ((nx != x || ny != y) && nx >= 0 && ny >= 0 && nx < WIDTH && ny < HEIGHT) {
                        break;
                    }
                }
                oldColor = get(nx, ny);
                newColor = get(x, y);
                set(nx, ny, newColor);
                debugPrint("Friend Lonely Pixel: " + newColor);
            }

            colorCount[oldColor]--;
            colorCount[newColor]++;
            if (oldColor != 0 && colorCount[oldColor] == 0) {
                usedColors.remove(oldColor);
                unusedColors.add(oldColor);
            }
            somethingChanged = true;
        }
        
        return somethingChanged;
    }
}
