package com.brackeen.evolution.core.game;

import com.brackeen.evolution.core.Actor;
import com.brackeen.evolution.core.Evolution;
import com.brackeen.evolution.core.Scene;
import com.brackeen.evolution.core.util.MathUtil;
import com.brackeen.evolution.core.util.TextElement;
import playn.core.Font;
import playn.core.GroupLayer;
import playn.core.ImmediateLayer;
import playn.core.Layer;
import playn.core.Mouse;
import playn.core.PlayN;
import playn.core.Surface;
import pythagoras.f.Transform;

public class GameScene extends Scene {
    
    private static final int BULLET_DISTANCE_SQ = SpriteCreator.HEIGHT * SpriteCreator.HEIGHT;
    private static final int ALIEN_DISTANCE_SQ = (SpriteCreator.HEIGHT * 3 / 2) * (SpriteCreator.HEIGHT * 3 / 2);

    private enum State {
        PLAYING,
        LEVEL_WON, 
        LEVEL_LOST,
    }
    
    private AlienGenome[] genomes;
    private AlienArmada armada;
    private Spaceship spaceship;
    private State state = State.PLAYING;
    private boolean fastMode = false;
    
    public GameScene(final Evolution evolution, AlienGenome[] genomes, final int level) {
        this.genomes = genomes;
        armada = new AlienArmada(evolution.getScreenWidth(), evolution.getScreenHeight(), 0);
        
        for (int x = 0; x < 8; x++) {
            createAlien(genomes[0], x, 3);
            createAlien(genomes[0], x, 4);
            createAlien(genomes[1], x, 1);
            createAlien(genomes[1], x, 2);
        }
        createAlien(genomes[2], 2, 0);
        createAlien(genomes[2], 3, 0);
        createAlien(genomes[2], 4, 0);
        createAlien(genomes[2], 5, 0);
        
        spaceship = new Spaceship(armada, level);
        Layer.HasSize l = spaceship.layer();
        l.setTranslation(evolution.getScreenWidth() / 2, evolution.getScreenHeight() - l.height() * 2);
        addActor(spaceship);
        
        if (level > 1) {
            Font font = PlayN.graphics().createFont("prstartk", Font.Style.PLAIN, 8 * Evolution.SCALE);
            TextElement label = new TextElement(font, "LEVEL " + level);
            label.setColor(Evolution.COLORS[1]);
            Layer.HasSize layer = label.layer();
            layer.setTranslation(5, 5);
            rootLayer().add(layer);
        }
        
        Layer.HasSize clickLayer = PlayN.graphics().createImmediateLayer(evolution.getScreenWidth(), evolution.getScreenHeight(), new ImmediateLayer.Renderer() {

            @Override
            public void render(Surface surface) {
                
            }
        });
        rootLayer().add(clickLayer);
        clickLayer.addListener(new Mouse.LayerAdapter() {
            @Override
            public void onMouseDown(Mouse.ButtonEvent event) {
                if (state == State.PLAYING) {
                    fastMode = !fastMode;
                }
                else if (state == State.LEVEL_WON) {
                    evolution.setCurrentScene(new SetupScene(evolution, GameScene.this.genomes, level + 1));
                }
                else {
                    evolution.setCurrentScene(new SetupScene(evolution, null, 1));
                }
            }
        });
        
        
        System.out.println(genomes[0]);
        System.out.println(genomes[1]);
        System.out.println(genomes[2]);
    }
   
    private void createAlien(AlienGenome genome, int x, int y) {
        Alien alien = new Alien(genome);
        alien.setArmadaPosition(x, y);
        addActor(alien);
    }
    
    @Override
    public void addActor(final Actor actor) {
        if (actor instanceof Alien) {
            armada.addAlien((Alien)actor);
        }
        super.addActor(actor);
    }
    
    @Override
    public void removeActor(final Actor actor) {
        if (actor instanceof Alien) {
            armada.removeAlien((Alien)actor);
        }
        super.removeActor(actor);
    }
    
    @Override
    public void update(Evolution evolution, float delta) {
        super.update(evolution, delta);
        armada.update(evolution, delta);
        
        if (state != State.PLAYING) {
            return;
        }
        
        
        int numAttacking = armada.getNumAliensInState(Alien.State.ATTACKING);
        int numRetreating = armada.getNumAliensInState(Alien.State.MOVING_TO_FORMATION);
        if (numAttacking == 0 && numRetreating == 0) {
            if (spaceship.shouldRemove()) {
                state = State.LEVEL_WON;
                addEndGameLabel(evolution, 
                        "MISSION COMPLETE\n\n" +
                        "CLICK TO CONTINUE",
                        Evolution.COLORS[1]);
                
                Evolution.playSound("audio/win");

                return;
            }
            else if (armada.getNumAliens() == 0) {
                state = State.LEVEL_LOST;
                addEndGameLabel(evolution, 
                        "MISSION FAILURE\n\n" +
                        "CLICK TO CONTINUE",
                        Evolution.COLORS[7]);
                Evolution.playSound("audio/lose");
                return;
            }
            
            int maxAttacking = Math.min(3, armada.getNumAliens());
            for (int i = numAttacking; i < maxAttacking; i++) {
                Alien alien = armada.getMostRestedAlien();
                if (alien != null) {
                    alien.attack(spaceship);
                }
            }
        }
        // Actor->Spaceship collisions
        if (!spaceship.shouldRemove()) {
            for (Actor actor : this) {
                if (!actor.shouldRemove()) {
                    boolean isAlien = false;
                    float damage = 50;
                    float maxDist = ALIEN_DISTANCE_SQ;
                    if (actor instanceof Alien) {
                        Alien alien = (Alien)actor;
                        isAlien = true;
                        damage = alien.getGenome().getShields();
                    }
                    else if (actor instanceof Bullet) {
                        Bullet bullet = (Bullet)actor;
                        if (bullet.getSource() == Bullet.Source.ALIEN) {
                            isAlien = true;
                            damage = bullet.getDamage();
                            maxDist = BULLET_DISTANCE_SQ;
                        }
                    }

                    if (isAlien) {
                        Transform t1 = spaceship.layer().transform();
                        Transform t2 = actor.layer().transform();
                        float distSq = MathUtil.distSq(t1.tx(), t1.ty(), t2.tx(), t2.ty());
                        if (distSq <= maxDist) {
                            spaceship.doDamage(damage);
                            actor.setShouldRemove(true);
                        }
                    }
                }
            }
        }
        
        // Spaceship Bullet -> Alien collisions
        for (Actor actor : this) {
            if (!actor.shouldRemove() && actor instanceof Bullet) {
                Bullet bullet = (Bullet)actor;
                if (bullet.getSource() == Bullet.Source.SPACESHIP) {
                    for (Alien alien : armada) {
                        if (!alien.shouldRemove()) {
                            Transform t1 = bullet.layer().transform();
                            Transform t2 = alien.layer().transform();
                            float distSq = MathUtil.distSq(t1.tx(), t1.ty(), t2.tx(), t2.ty());
                            if (distSq <= BULLET_DISTANCE_SQ) {
                                alien.doDamage(bullet.getDamage());
                                bullet.setShouldRemove(true);
                            }
                        }
                    }
                }
            }
        }
    }
    
    @Override
    public int getUpdateIterations() {
        if (state == State.PLAYING && fastMode) {
            return 3;
        }
        else {
            return 1;
        }
    }
    
    private void addEndGameLabel(final Evolution evolution, String text, int color) {
        Font font = PlayN.graphics().createFont("prstartk", Font.Style.PLAIN, 8 * Evolution.SCALE);
        Layer.HasSize layer;
        
        TextElement label = new TextElement(font, text);
        label.setColor(color);
        layer = label.layer();
        layer.setOrigin(layer.width()/2, layer.height()/2);
        layer.setTranslation(evolution.getScreenWidth()/2, evolution.getScreenHeight()/2);
        rootLayer().add(layer);
        
    }
}
