package com.brackeen.evolution.core.game;

import com.brackeen.evolution.core.Actor;
import com.brackeen.evolution.core.Evolution;
import com.brackeen.evolution.core.Scene;
import playn.core.GroupLayer;
import playn.core.Image;
import playn.core.ImageLayer;
import playn.core.ImmediateLayer;
import playn.core.Layer;
import playn.core.PlayN;
import playn.core.Surface;

public class Spaceship extends Actor {
    
    private static final boolean SHOW_AI_GOAL = false;
    
    private static final int[][] IMAGE = {
        { 0, 0, 0, 0, 0, 0, 0, 1, },
        { 0, 0, 0, 0, 0, 0, 0, 1, },
        { 0, 0, 0, 0, 0, 0, 1, 1, },
        { 0, 0, 0, 0, 0, 0, 1, 1, },
        { 0, 0, 0, 0, 0, 0, 1, 1, },
        { 0, 0, 0, 0, 0, 0, 1, 1, },
        { 0, 0, 0, 2, 0, 0, 1, 1, },
        { 0, 0, 0, 2, 0, 0, 1, 1, },
        { 2, 0, 0, 1, 0, 1, 1, 1, },
        { 2, 0, 0, 1, 1, 1, 1, 3, },
        { 1, 0, 0, 1, 1, 1, 3, 3, },
        { 1, 0, 1, 1, 1, 1, 3, 1, },
        { 1, 1, 1, 1, 1, 1, 1, 1, },
        { 1, 1, 1, 3, 3, 3, 1, 1, },
        { 1, 1, 0, 0, 0, 0, 0, 1, },
        { 1, 0, 0, 0, 0, 0, 0, 1, },
    };
    
    private final AlienArmada armada;
    
    // Stats
    private final float maxHealth;
    private final float velocity;
    private final float fireDamage;
    private final float bulletSpeed;
    private float health;
    
    // AI
    private boolean idle;
    private float goalX;
    private float goalY;
    private float durationSinceDecision;
    private Layer.HasSize debugAiBar;
    
    // Firing
    private Image fireImage;
    private final float timeBetweenFiring;
    private float timeUntilCanFire;
    
    // Health bar 
    private Layer.HasSize healthBar;
    private float flashHealthBar;
    
    public Spaceship(AlienArmada armada, int level) {
        
        this.armada = armada;
        
        Image image = SpriteCreator.create(IMAGE);
        ImageLayer layer = PlayN.graphics().createImageLayer(image);
        layer.setOrigin(image.width()/2, image.height()/2);
        setLayer(layer);
        
        maxHealth = 200 + (100 * level);
        health = maxHealth;
        velocity = layer.width() * 3 + layer.width() * Math.min(4, level - 1);
        bulletSpeed = layer.width() * 12;
        
        fireDamage = 50 + (level - 1) * 5;
        timeBetweenFiring = Math.max(0.4f, 0.8f - (level - 1) * 0.05f);
        
        
        fireImage = SpriteCreator.create(Bullet.SPACESHIP);
        
        // Wait a bit
        idle = true;
        timeUntilCanFire = 2.0f;
    }

    public void doDamage(float damage) {
        health -= damage;
        flashHealthBar = 0.100f;
        
        if (health <= 0) {
            if (healthBar != null) {
                GroupLayer group = layer().parent();
                group.remove(healthBar);
                healthBar = null;
            }
            setShouldRemove(true);
            Evolution.playSound("audio/explode");
        }
        else {
            Evolution.playSound("audio/hit");
        }
    }
    
    @Override
    public void update(Evolution evolution, float delta) {
        final float elapsedTime = delta / 1000.0f;
        
        durationSinceDecision += elapsedTime;
        
        if (healthBar == null && health > 0) {
            GroupLayer group = layer().parent();
            float w = layer().width();
            float h = layer().height() / 8;
            healthBar = PlayN.graphics().createImmediateLayer((int)w, (int)h, new ImmediateLayer.Renderer() {

                @Override
                public void render(Surface surface) {
                    if (healthBar != null && health > 0) {
                        float p = health / maxHealth;
                        if (flashHealthBar > 0) {
                            surface.setFillColor(Evolution.COLORS[1]);
                            flashHealthBar -= elapsedTime;
                        }
                        else if (p > 0.5) {
                            surface.setFillColor(Evolution.COLORS[4]);
                        }
                        else if (p > 0.25) {
                            surface.setFillColor(Evolution.COLORS[6]);
                        }
                        else {
                            surface.setFillColor(Evolution.COLORS[2]);
                        }
                        float w = healthBar.width() * p;
                        float mw = Math.max(1, w);
                        surface.fillRect((healthBar.width() - w) / 2, 0, mw, healthBar.height());
                    }
                }
            });
            healthBar.setOrigin(w/2, h/2);
            group.add(healthBar);
        }
                
        if (idle) {
            if (armada.getNumAliensInState(Alien.State.MOVING_TO_FORMATION) == 0) {
                idle = false;
                decideAI(evolution);
            }
        }
        else {
            // Move to goal position
            float x = layer().transform().tx();
            float y = layer().transform().ty();
            
            boolean goalXMet = false;
            boolean goalYMet = false;
            
            float dist = velocity * elapsedTime;
            if (Math.abs(x - goalX) < dist) {
                x = goalX;
                goalXMet = true;
            }
            else if (x < goalX) {
                x += dist;
            }
            else {
                x -= dist;
            }
            if (Math.abs(y - goalY) < dist) {
                y = goalY;
                goalYMet = true;
            }
            else if (y < goalY) {
                y += dist;
            }
            else {
                y -= dist;
            }
            layer().setTranslation(x, y);
            if (goalXMet && goalYMet) {
                decideAI(evolution);
            }
            else if (durationSinceDecision > 0.5f) {
                // Force decision
                decideAI(evolution);
            }
        }
        
        // Fire
        if (timeUntilCanFire <= 0) {
            float x = layer().transform().tx();
            float y = layer().transform().ty();
            float dist = layer().width() / 2;
            boolean alienCloseBy = false;
            for (Alien alien : armada) {
                float ax = alien.layer().transform().tx();
                float ay = alien.layer().transform().ty();
                if (Math.abs(ax - x) <= dist && ay < y) {
                    alienCloseBy = true;
                    break;
                }
            }
            
            if (alienCloseBy) {
                Scene scene = getScene();
                if (scene != null) {
                    Bullet bullet = new Bullet(Bullet.Source.SPACESHIP, fireImage, x, y, -bulletSpeed, fireDamage);
                    scene.addActor(bullet);
                    timeUntilCanFire = timeBetweenFiring;
                    Evolution.playSound("audio/shoot");
                }
            }
        }
        else {
            timeUntilCanFire -= elapsedTime;
        }
                
        if (healthBar != null) {
            float hx = layer().transform().tx() - Evolution.SCALE/2;
            float hy = layer().transform().ty() + layer().height() * 3 / 4;
            healthBar.setTranslation((int)hx, (int)hy);
        }
    }
    
    private void decideAI(Evolution evolution) {
        float x = layer().transform().tx();
        float y = layer().transform().ty();
        goalX = x;
        goalY = y;
        durationSinceDecision = 0;
            
        final int w = 32;
        boolean attackingAlienFound = false;
        boolean[] attackingAliens = new boolean[w];
        Alien closestAlien = null;
        for (Alien alien : armada) {
            if (closestAlien == null) {
                closestAlien = alien;
            }
            else {
                float ax = alien.layer().transform().tx();
                float cax = closestAlien.layer().transform().tx();
                if (Math.abs(x - ax) < Math.abs(x - cax)) {
                    closestAlien = alien;
                }
            }
            if (alien.getState() == Alien.State.ATTACKING &&  armada.isPastFrontLine(alien)) {
                float ax = alien.layer().transform().tx();
                float ay = alien.layer().transform().ty();
                int i = Math.round(ax * w / evolution.getScreenWidth());
                if (i >= 0 && i < w) {
                    attackingAliens[i] = true;
                    attackingAlienFound = true;
                    
                    if (Math.abs(ay - y) < layer().height() * 4) {
                        // Very close
                        if (i > 0) {
                            attackingAliens[i-1] = true;
                        }
                        if (i < w - 1) {
                            attackingAliens[i+1] = true;
                        }
                    }
                }
            }
        }
        
        Scene scene = getScene();
        if (scene != null) {
            for (Actor actor : scene) {
                if (actor instanceof Bullet) {
                    Bullet bullet = (Bullet)actor;
                    if (bullet.getSource() == Bullet.Source.ALIEN) {
                        float ax = bullet.layer().transform().tx();
                        int i = Math.round(ax * w / evolution.getScreenWidth());
                        if (i >= 0 && i < w) {
                            attackingAliens[i] = true;
                            attackingAlienFound = true;
                        }
                    }
                }
            }
        }
        boolean playerInHole = false;
        int pi = Math.round(x * w / evolution.getScreenWidth());
        if (pi >= 0 && pi < w) {
            playerInHole = !attackingAliens[pi];
        }
        
        if (attackingAlienFound && !playerInHole) {
            // Find largest "hole", go to the middle of it.
            int longestRun = -1;
            int longestRunLength = 0;
            int currentRun = -1;
            boolean inRun = false;
            for (int i = 0; i < w; i++) {
                if (!inRun) {
                    if (attackingAliens[i] == false) {
                        inRun = true;
                        currentRun = i;
                    }
                }
                else {
                    if (attackingAliens[i]) {
                        int runLength = i - currentRun;
                        if (runLength > longestRunLength) {
                            longestRun = currentRun;
                            longestRunLength = runLength;
                        }
                        inRun = false;
                    }
                }
            }
            
            if (longestRun != -1) {
                boolean found = false;
                float minX = longestRun * evolution.getScreenWidth() / w;
                float maxX = (longestRun + longestRunLength) * evolution.getScreenWidth() / w;
                
//                if (closestAlien != null) {
//                    float cax = closestAlien.layer().transform().tx();
//                    if (cax > minX && cax < maxX) {
//                        goalX = cax;
//                        found = true;
//                    }
//                }
                
                if (!found) {
                    float p = longestRun + longestRunLength / 2.0f;
                    goalX = p * evolution.getScreenWidth() / w;
                }
                //System.out.println(Arrays.toString(attackingAliens));
                //System.out.println("Goto: " + longestRun + " " + longestRunLength);
            }
        }
        else if (closestAlien != null) {
            // Find closest alien (horizontally) and go there
            goalX = closestAlien.layer().transform().tx();
        }
        
        if (SHOW_AI_GOAL) {
            if (debugAiBar == null) {
                GroupLayer group = layer().parent();
                final int dw = 8;
                final int dh = 8;
                debugAiBar = PlayN.graphics().createImmediateLayer(dw, dh, new ImmediateLayer.Renderer() {

                    @Override
                    public void render(Surface surface) {
                        surface.setFillColor(Evolution.COLORS[2]);
                        surface.fillRect(0, 0, dw, dh);
                    }
                });
                debugAiBar.setOrigin(dw/2, dh/2);
                group.add(debugAiBar);
            }
            debugAiBar.setTranslation(goalX, goalY + layer().height());
        }
    }
}
