package com.brackeen.evolution.core.util;

/* 
This code is from the BSD-license PulpCore framework.
*/
public class BezierPath {

    private static final float PI = (float)Math.PI;
    private static final float TWO_PI = (float)(Math.PI * 2);

    /**
        The move command. The move-to command requires one point: the location
        to move to.
    */
    public static final int MOVE_TO = 0;

    /**
        The line segment command. The line-to command requires one point:
        the location to draw a line segment to.
    */
    public static final int LINE_TO = 1;

    /**
        The cubic bezier command. The curve-to command requires three points.
        The first point is the control point at the beginning of the curve,
        the second point is the control point at the end of the curve,
        and the third point is the final destination point. All points are
        absolute values.
    */
    public static final int CURVE_TO = 2;

    /** The number of points in the path. */
    private final int numPoints;

    /** The x-location of each point. */
    private final float[] xPoints;

    /** The y-location of each point. */
    private final float[] yPoints;

    /** The value, from 0 to 1, representing the portion of the total length at each point. */
    private final float[] pPoints;

    private float totalLength;

    private boolean isClosed;

    private transient float lastCalcP = -1;
    private transient float[] lastCalcPoint = new float[2];

    public BezierPath(float[]... commands)  {

        float[][] points = toLineSegments(commands);

        this.xPoints = points[0];
        this.yPoints = points[1];
        this.numPoints = xPoints.length;
        this.pPoints = new float[numPoints];

        init();
    }

    /**
        Returns true if the path is closed, that is, the first point and the last point
        are identical.
     */
    public boolean isClosed() {
        return isClosed;
    }

    public float getLength() {
        return totalLength;
    }

    private void init() {

        // Find the lengths of each segment;
        totalLength = 0;
        float[] segmentLength = new float[numPoints - 1];
        for (int i = 0; i < numPoints - 1; i++) {
            float dist = MathUtil.dist(xPoints[i], yPoints[i], xPoints[i + 1], yPoints[i + 1]);
            segmentLength[i] = dist;
            totalLength += dist;
        }

        pPoints[0] = 0;
        pPoints[numPoints - 1] = 1;
        float length = 0;
        for (int i = 1; i < numPoints - 1; i++) {
            length += segmentLength[i - 1];
            pPoints[i] = length / totalLength;
        }

        isClosed = xPoints[0] == xPoints[numPoints - 1] && yPoints[0] == yPoints[numPoints - 1];
    }

    public void translate(float x, float y) {
        if (x == 0 && y == 0) {
            return;
        }
        for (int i = 0; i < numPoints; i++) {
            xPoints[i] += x;
            yPoints[i] += y;
        }
    }

    public float getStartX() {
        return xPoints[0];
    }

    public float getStartY() {
        return yPoints[0];
    }

    public float getEndX() {
        return xPoints[numPoints - 1];
    }

    public float getEndY() {
        return yPoints[numPoints - 1];
    }

    /**
        Gets the x location of point p on the path, where p is typically from 0 (start of the path)
        to 1 (end of the path).
        @param p The position along the path to place the sprite, from 0 to 1.
     */
    public float getX(float p) {
        float px = clampP(p);
        return get(0, px);
    }

    /**
        Gets the y location of point p on the path, where p is typically from 0 (start of the path)
        to 1 (end of the path).
        @param p The position along the path to place the sprite, from 0 to 1.
     */
    public float getY(float p) {
        float px = clampP(p);
        return get(1, px);
    }

    /**
        Gets the angle of point p on the path, where p is typically from 0 (start of the path)
        to 1 (end of the path).
        @param p The position along the path to place the sprite, from 0 to 1.
     */
    public float getAngle(float p) {
        float px = clampP(p);
        return get(2, px);
        //int i = getLowerSegment(px);
        //return Math.atan2(yPoints[i + 1] - yPoints[i], xPoints[i + 1] - xPoints[i]);
    }

    private float clampP(float p) {
        if (isClosed()) {
            // Only wrap if the path is closed.
            // Special case: ONE is considered the end.
            if (p != 1) {
                p %= 1;
                if (p < 0) {
                    p += 1;
                }
            }
        }
        else {
            p = MathUtil.clamp(p, 0, 1);
        }
        return p;
    }

    private float get(int axis, float p) {
        float px = clampP(p);
        if (axis == 0 || axis == 1) {
            return getLocation(px)[axis];
        }
        else {
            int i = getLowerSegment(px);
            float lowerAngle = (float)Math.atan2(yPoints[i + 1] - yPoints[i],
                    xPoints[i + 1] - xPoints[i]);
            if (i == numPoints - 2) {
                return lowerAngle;
            }
            else {
                float higherAngle = (float)Math.atan2(yPoints[i + 2] - yPoints[i + 1],
                        xPoints[i + 2] - xPoints[i + 1]);
                float dAngle = higherAngle - lowerAngle;
                if (Math.abs(dAngle + TWO_PI) < Math.abs(dAngle)) {
                    dAngle += TWO_PI;
                }
                else if (Math.abs(dAngle - TWO_PI) < Math.abs(dAngle)) {
                    dAngle -= TWO_PI;
                }

                return lowerAngle + dAngle * (px - pPoints[i]) /  (pPoints[i + 1] - pPoints[i]);
            }
        }
    }

    private int getLowerSegment(float p) {

        // Binary search for P
        int high = numPoints - 1;
        int low = 0;
        while (high - low > 1) {
            int index = (high + low) >>> 1;

            if (pPoints[index] > p) {
                high = index;
            }
            else {
                low = index;
            }
        }

        return low;
    }

    private float[] getLocation(float p) {

        if (p == lastCalcP) {
            return lastCalcPoint;
        }

        lastCalcP = p;

        if (p == 0) {
            lastCalcPoint[0] = xPoints[0];
            lastCalcPoint[1] = yPoints[0];
            return lastCalcPoint;
        }
        else if (p == 1) {
            lastCalcPoint[0] = xPoints[numPoints - 1];
            lastCalcPoint[1] = yPoints[numPoints - 1];
            return lastCalcPoint;
        }

        // Binary search for P
        int high = numPoints - 1;
        int low = 0;
        while (high - low > 1) {
            int index = (high + low) >>> 1;

            if (pPoints[index] > p) {
                high = index;
            }
            else {
                low = index;
            }
        }

        if (high == low) {
            lastCalcPoint[0] = xPoints[low];
            lastCalcPoint[1] = yPoints[low];
        }
        else {
            float q = p - pPoints[low];
            float r = pPoints[high] - pPoints[low];

            lastCalcPoint[0] = xPoints[low] + (xPoints[high] - xPoints[low]) * q / r;
            lastCalcPoint[1] = yPoints[low] + (yPoints[high] - yPoints[low]) * q / r;
        }

        return lastCalcPoint;
    }

    //
    // Convert draw commands to line segments.
    //

    private static float[][] toLineSegments(float[]... drawCommands) {

        int numDrawCommands = drawCommands.length;
        float[][][] allPoints = new float[numDrawCommands][2][];
        int numPoints = 0;
        float x = 0;
        float y = 0;

        for (int i = 0; i < numDrawCommands; i++) {
            float[] drawCommand = drawCommands[i];
            switch ((int)drawCommand[0]) {

                case MOVE_TO: case LINE_TO:
                    x = drawCommand[1];
                    y = drawCommand[2];

                    allPoints[i] = new float[2][1];
                    allPoints[i][0][0] = x;
                    allPoints[i][1][0] = y;
                    numPoints++;
                    break;

                case CURVE_TO:
                    allPoints[i] = toLineSegments(x, y,
                         drawCommand[1],  drawCommand[2],
                         drawCommand[3],  drawCommand[4],
                         drawCommand[5],  drawCommand[6]);

                    x = drawCommand[5];
                    y = drawCommand[6];
                    numPoints += allPoints[i][0].length;
                    break;
            }
        }

        // Convert the allPoints array to xPoints, yPoints
        float[] xPoints = new float[numPoints];
        float[] yPoints = new float[numPoints];
        int offset = 0;
        for (int i = 0; i < numDrawCommands; i++) {
            int len = allPoints[i][0].length;
            System.arraycopy(allPoints[i][0], 0, xPoints, offset, len);
            System.arraycopy(allPoints[i][1], 0, yPoints, offset, len);

            offset += len;
            allPoints[i] = null;
        }

        float[][] returnValue = new float[2][];
        returnValue[0] = xPoints;
        returnValue[1] = yPoints;

        return returnValue;
    }

    //
    // Bezier conversion to line segments (using forward differencing).
    //

    private static float[][] toLineSegments(float x1, float y1, float x2, float y2,
        float x3, float y3, float x4, float y4)
    {
        // First division
        float x12   = (x1 + x2) / 2f;
        float y12   = (y1 + y2) / 2f;
        float x23   = (x2 + x3) / 2f;
        float y23   = (y2 + y3) / 2f;
        float x34   = (x3 + x4) / 2f;
        float y34   = (y3 + y4) / 2f;
        float x123  = (x12 + x23) / 2f;
        float y123  = (y12 + y23) / 2f;
        float x234  = (x23 + x34) / 2f;
        float y234  = (y23 + y34) / 2f;
        float x1234 = (x123 + x234) / 2f;
        float y1234 = (y123 + y234) / 2f;

        // Left division
        float lx12   = (x1 + x12) / 2f;
        float ly12   = (y1 + y12) / 2f;
        float lx23   = (x12 + x123) / 2f;
        float ly23   = (y12 + y123) / 2f;
        float lx34   = (x123 + x1234) / 2f;
        float ly34   = (y123 + y1234) / 2f;
        float lx123  = (lx12 + lx23) / 2f;
        float ly123  = (ly12 + ly23) / 2f;
        float lx234  = (lx23 + lx34) / 2f;
        float ly234  = (ly23 + ly34) / 2f;
        float lx1234 = (lx123 + lx234) / 2f;
        float ly1234 = (ly123 + ly234) / 2f;

        // Right division
        float rx12   = (x1234 + x234) / 2f;
        float ry12   = (y1234 + y234) / 2f;
        float rx23   = (x234 + x34) / 2f;
        float ry23   = (y234 + y34) / 2f;
        float rx34   = (x34 + x4) / 2f;
        float ry34   = (y34 + y4) / 2f;
        float rx123  = (rx12 + rx23) / 2f;
        float ry123  = (ry12 + ry23) / 2f;
        float rx234  = (rx23 + rx34) / 2f;
        float ry234  = (ry23 + ry34) / 2f;
        float rx1234 = (rx123 + rx234) / 2f;
        float ry1234 = (ry123 + ry234) / 2f;

        // Determine the number of segments for each division
        int numSegments1 = getNumSegments(x1, y1, lx12, ly12, lx123, ly123, lx1234, ly1234);
        int numSegments2 = getNumSegments(lx1234, ly1234, lx234, ly234, lx34, ly34, x1234, y1234);
        int numSegments3 = getNumSegments(x1234, y1234, rx12, ry12, rx123, ry123, rx1234, ry1234);
        int numSegments4 = getNumSegments(rx1234, ry1234, rx234, ry234, rx34, ry34, x4, y4);

        int totalPoints = numSegments1 + numSegments2 + numSegments3 + numSegments4;
        float[] xPoints = new float[totalPoints];
        float[] yPoints = new float[totalPoints];
        int offset = 0;

        // Convert to lines
        offset += toLineSegments(xPoints, yPoints, offset, numSegments1,
            x1, y1, lx12, ly12, lx123, ly123, lx1234, ly1234);
        offset += toLineSegments(xPoints, yPoints, offset, numSegments2,
            lx1234, ly1234, lx234, ly234, lx34, ly34, x1234, y1234);
        offset += toLineSegments(xPoints, yPoints, offset, numSegments3,
            x1234, y1234, rx12, ry12, rx123, ry123, rx1234, ry1234);
        offset += toLineSegments(xPoints, yPoints, offset, numSegments4,
            rx1234, ry1234, rx234, ry234, rx34, ry34, x4, y4);

        float[][] returnValue = new float[2][];
        returnValue[0] = xPoints;
        returnValue[1] = yPoints;
        return returnValue;
    }

    private static int getNumSegments(float x0, float y0, float x1, float y1, float x2, float y2,
        float x3, float y3)
    {
        int numSegments;

        // attempt to measure the "curvature" of the segment - more curvature, the more points.
        /*
        float dx = (x1 + x2 - x0 - x3) / 2;
        float dy = (y1 + y2 - y0 - y3) / 2;

        int numSegments;
        if (dx == 0 && dy == 0) {
            numSegments = 1;
        }
        else {
            float distSq = dx * dx + dy * dy;

            int log2Dist = MathUtil.log2(Math.round(distSq)) >> 1;

            numSegments = Math.max(1, 1 << (log2Dist - 1));
        }
        */

        float dist = Math.max(
            getDist(x1, y1, x0, y0, x3, y3),
            getDist(x2, y2, x0, y0, x3, y3));

        if (dist <= 0) {
            numSegments = 1;
        }
        else {
            numSegments = Math.max(1, 1 << (MathUtil.log2(Math.round(dist))));
            numSegments = Math.min(numSegments, 16);
        }

        return numSegments;
    }

    private static int toLineSegments(float[] xPoints, float[] yPoints, int offset, int numSegments,
        float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3)
    {
        float t = 1f / numSegments;
        float tSquared = t * t;
        float tCubed = tSquared * t;

        float xf = x0;
        float xfd = 3 * (x1 - x0) * t;
        float xfdd2 = 3 * (x0 - 2 * x1 + x2) * tSquared;
        float xfddd6 = (3 * (x1 - x2) + x3 - x0) * tCubed;
        float xfddd2 = 3 * xfddd6;
        float xfdd = 2 * xfdd2;
        float xfddd = 2 * xfddd2;

        float yf = y0;
        float yfd = 3 * (y1 - y0) * t;
        float yfdd2 = 3 * (y0 - 2 * y1 + y2) * tSquared;
        float yfddd6 = (3 * (y1 - y2) + y3 - y0) * tCubed;
        float yfddd2 = 3 * yfddd6;
        float yfdd = 2 * yfdd2;
        float yfddd = 2 * yfddd2;

        for (int i = 1; i < numSegments; i++) {
            xf += xfd + xfdd2 + xfddd6;
            xfd += xfdd + xfddd2;
            xfdd += xfddd;
            xfdd2 += xfddd2;

            yf += yfd + yfdd2 + yfddd6;
            yfd += yfdd + yfddd2;
            yfdd += yfddd;
            yfdd2 += yfddd2;

            xPoints[offset] = xf;
            yPoints[offset] = yf;
            offset++;
        }

        xPoints[offset] = x3;
        yPoints[offset] = y3;
        offset++;

        return numSegments;
    }

    private static float getDist(float px, float py, float ax, float ay, float bx, float by) {

        // Distance from a point to a line approximation. From Graphics Gems II, page 11.

        float dx = Math.abs(bx - ax);
        float dy = Math.abs(by - ay);

        float div = dx + dy - Math.min(dx, dy)/2;

        if (div == 0) {
            return 0;
        }

        float a2 = (py - ay) * (bx - ax) - (px - ax) * (by - ay);

        return Math.abs(a2) / div;
    }
}